import os
import shutil
import subprocess
import sys
import textwrap
import traceback
from abc import ABC
from pathlib import Path
from typing import List

import yaml

from hamstercage.hamstercage_exception import HamstercageException
from hamstercage.utils import chmod, mkdir_with_owner_group_mode, path_as_child_of

"""
Manifest of files to be managed.
"""


class FileMode(int):
    pass


class Entry(ABC):
    mode: int
    group: str
    owner: str
    path: str
    target: str

    def __init__(self, path):
        self.path = path
        self.form = "normal"
        self.mode = 0o644
        self.owner = "root"
        self.group = "root"
        self.target = ""

    @staticmethod
    def entry(path: str, e):
        """
        Create a new entry.
        :param path: the relative path of the entry,
        :param e: a dict or Path describing the file to be added
        :return:
        """
        if isinstance(e, Path):
            if e.is_symlink():
                # is_dir() and is_file return true if the target is dir or file, so check for symlink first
                return SymlinkEntry.from_dict(path, {"target": str(e.resolve())})
            elif e.is_dir():
                return DirEntry.from_dict(
                    path,
                    {
                        "mode": e.stat().st_mode
                        & 0o7777,  # limit to standard POSIX bits
                        "owner": e.owner(),
                        "group": e.group(),
                    },
                )
            elif e.is_file():
                return FileEntry.from_dict(
                    path,
                    {
                        "target": e,
                        "mode": e.stat().st_mode
                        & 0o7777,  # limit to standard POSIX bits
                        "owner": e.owner(),
                        "group": e.group(),
                    },
                )
            raise HamstercageException(
                f'Unable to create entry for "{path}": unknown file type'
            )

        t = e.get("type", "file")
        if t == "dir":
            return DirEntry.from_dict(path, e)
        elif t == "file":
            return FileEntry.from_dict(path, e)
        elif t == "link":
            return SymlinkEntry.from_dict(path, e)
        else:
            raise HamstercageException(f'Unknown entry type "{e["type"]}"')

    def has_repo(self):
        """
        Returns true if this entry has a file in repo
        :return:
        """
        return False

    def apply(self, repo: Path, target: Path):
        """
        Apply the entry to the target dir.
        :param repo: Repo base dir
        :param target: Target base dir
        :return:
        """
        raise HamstercageException(f"class {self} does not implement apply()")

    def save(self, repo_path, target_path, manifest):
        raise RuntimeError("Internal error: cannot call abstract base class method")

    def path_as_child_of(self, target_path: Path) -> Path:
        """
        Returns the entry path as a child of target_path.
        :param target_path: the base path
        :return: path
        """
        return path_as_child_of(self.path, target_path)


class DirEntry(Entry):
    def __init__(self, path):
        super().__init__(path)

    @staticmethod
    def from_dict(path: str, d: dict) -> "DirEntry":
        e = DirEntry(path)
        if "mode" in d:
            e.mode = d["mode"]
        if "owner" in d:
            e.owner = d["owner"]
        if "group" in d:
            e.group = d["group"]
        if isinstance(e.mode, str):
            e.mode = int(e.mode, 8)
        e.path = Manifest.normalize_path(e.path)
        return e

    def to_dict(self):
        """
        Returns a dict representing this entry.
        :return:
        """
        return {
            "type": "dir",
            "mode": f"{self.mode:#o}",
            "owner": self.owner,
            "group": self.group,
        }

    def apply(self, repo: Path, target: Path):
        if target.exists() and not target.is_dir():
            raise HamstercageException(
                f'Unable to update "{target}" because it exists and is not a directory'
            )
        target.mkdir(self.mode, exist_ok=True, parents=True)
        shutil.chown(str(target), self.owner, self.group)

    def save(self, repo_path, target_path, manifest):
        if target_path.exists() and not target_path.is_dir:
            raise HamstercageException(
                f'Unable to update "{target_path}" because it exists and is not a directory'
            )
        self.group = target_path.group()
        self.mode = target_path.stat().st_mode & 0o7777
        self.owner = target_path.owner()

    def __str__(self):
        return f"SymlinkEntry<form={self.form}, path={self.path}, target={self.target}>"


class FileEntry(Entry):
    def __init__(self, path):
        super().__init__(path)

    @staticmethod
    def from_dict(path: str, d: dict) -> "FileEntry":
        e = FileEntry(path)
        if "target" in d:
            e.target = d["target"]
        if "mode" in d:
            e.mode = d["mode"]
        if "owner" in d:
            e.owner = d["owner"]
        if "group" in d:
            e.group = d["group"]
        if isinstance(e.mode, str):
            e.mode = int(e.mode, 8)
        e.path = Manifest.normalize_path(e.path)
        return e

    def to_dict(self):
        """
        Returns a dict representing this entry.
        :return:
        """
        return {
            "type": "file",
            "mode": f"{self.mode:#o}",
            "owner": self.owner,
            "group": self.group,
        }

    def has_repo(self):
        """
        Returns true if this entry has a file in repo
        :return:
        """
        return True

    def apply(self, repo: Path, target: Path):
        if target.exists() and not target.is_file():
            raise HamstercageException(
                f'Unable to update "{target}" because it exists and is not a file'
            )
        target.parent.mkdir(exist_ok=True, parents=True)
        shutil.copy2(str(repo), str(target))
        chmod(str(target), self.mode)
        shutil.chown(str(target), self.owner, self.group)

    def save(self, repo_path, target_path, manifest):
        if repo_path.exists() and not repo_path.is_file():
            raise HamstercageException(
                f"Unable to add {repo_path}: another directory entry already exists here"
            )
        try:
            self.group = target_path.group()
            self.mode = target_path.stat().st_mode & 0o7777
            self.owner = target_path.owner()
            mkdir_with_owner_group_mode(
                repo_path.parent, manifest.owner, manifest.group, manifest.dir_mode
            )
            shutil.copy2(target_path, repo_path, follow_symlinks=False)
            shutil.chown(repo_path, manifest.owner, manifest.group)
            repo_path.chmod(manifest.file_mode)
        except FileNotFoundError as e:
            raise HamstercageException(
                f"Unable to save {repo_path} to tag {target_path}: {e}", e
            )

    def __str__(self):
        return (
            f"FileEntry<form={self.form}, path={self.path}, mode={self.mode:#o}"
            + ", owner={self.owner}, group={self.group}>"
        )


class SymlinkEntry(Entry):
    def __init__(self, path):
        super().__init__(path)

    @staticmethod
    def from_dict(path: str, d: dict) -> "SymlinkEntry":
        e = SymlinkEntry(path)
        if not "target" in d:
            raise HamstercageException("missing target attribute for symlink")
        e.target = d["target"]
        if "mode" in d:
            e.mode = d["mode"]
        if "owner" in d:
            e.owner = d["owner"]
        if "group" in d:
            e.group = d["group"]
        if isinstance(e.mode, str):
            e.mode = int(e.mode, 8)
        e.path = Manifest.normalize_path(e.path)
        return e

    def to_dict(self):
        """
        Returns a dict representing this entry.
        :return:
        """
        return {
            "type": "link",
            "target": self.target,
        }

    def apply(self, repo: Path, target: Path):
        if target.exists() and not target.is_symlink():
            raise HamstercageException(
                f'Unable to update "{target}" because it exists and is not a symbolic link'
            )
        target.parent.mkdir(exist_ok=True, parents=True)
        if target.exists():
            target.unlink()
        target.symlink_to(self.target)

    def save(self, repo_path: Path, target_path: Path, manifest: "Manifest"):
        if target_path.exists() and not target_path.is_symlink():
            raise HamstercageException(
                f'Unable to save "{target_path}" because it is not a symbolic link'
            )
        self.target = str(os.readlink(target_path))

    def __str__(self):
        return f"SymlinkEntry<form={self.form}, path={self.path}, target={self.target}>"


class Host:
    """
    Represents one host in the manifest.
    """

    description: str
    name: str
    tags: List[str]

    def __init__(self, n: str) -> None:
        self.description = ""
        self.name = n
        self.tags = []
        pass

    @staticmethod
    def from_dict(n: str, d: dict) -> "Host":
        host = Host(n)
        if "description" in d:
            host.description = d.get("description", "")
        if not "tags" in d:
            raise HamstercageException(
                f'Invalid host definition for "{n}": must have a list of tags'
            )
        host.tags = d["tags"]
        return host

    def to_dict(self) -> dict:
        d = {"tags": self.tags}
        if len(self.description) > 0:
            d["description"] = self.description
        return d


class Hook:
    """
    Represents one hook in a tag.
    """

    command: str
    description: str
    name: str
    type: str

    valid_types = ["exec", "python", "shell"]

    def __init__(self, name: str):
        pass

    @staticmethod
    def from_dict(name: str, d: dict) -> "Hook":
        hook = Hook(name)
        if "command" not in d:
            raise HamstercageException(
                f'In definition of hook "{hook.name}": missing "command"'
            )
        hook.command = d["command"]
        hook.description = d.get("description", "")
        hook.name = name
        hook.type = d["type"]
        if hook.type not in Hook.valid_types:
            raise HamstercageException(
                f'In definition of hook "{hook.name}": Invalid hook type "{hook.type}", must be one of {", ".join(Hook.valid_types)}'
            )
        return hook

    def to_dict(self) -> dict:
        d = {
            "command": self.command,
            "description": self.description,
            "type": self.type,
        }
        return d

    def call(self, manifest: "Manifest", cmd: str, step: str, tag: "Tag") -> int:
        if self.type == "exec":
            return self._call_shell(
                [str(self._get_path(manifest)), cmd, step, tag.name],
                manifest,
                cmd,
                step,
                tag,
                shell=False,
            )
        elif self.type == "python":
            return self._call_python(manifest, cmd, step, tag)
        elif self.type == "shell":
            return self._call_shell(
                [self.command], manifest, cmd, step, tag, shell=True
            )
        else:
            raise HamstercageException(
                f'In definition of hook "{self.name}": Invalid hook type "{self.type}", must be one of {", ".join(Hook.valid_types)}'
            )

    def _call_python(self, manifest: "Manifest", cmd: str, step: str, tag: "Tag"):
        path = self._get_path(manifest)
        globals = {
            "cmd": cmd,
            "manifest": manifest,
            "hook": self.name,
            "repo": str(Path(manifest.manifest_file).parent),
            "step": step,
            "tag": tag,
            "__file__": str(path),
            "__name__": "__hamstercage__",
        }
        script = path.read_text("utf-8")
        try:
            exec(script, globals)
            return 0
        except SyntaxError as e:
            lines = script.split("\n")
            raise HamstercageException(
                f'Error executing hook "{self.name}" Python command "{self.command}" line {e.lineno}::\n\t{lines[e.lineno - 1]}'
            )
        except Exception as e:
            _, _, tb = sys.exc_info()
            tb_info = traceback.extract_tb(tb)
            filename, line, func, text = tb_info[-1]
            lines = script.split("\n")
            print(
                textwrap.dedent(
                    f"""
                            Error executing hook "{self.name}" Python command "{self.command}" line {line}: {e.__class__.__name__}{e.args}
                            \t{lines[line - 1]}
                            """
                ),
                file=sys.stderr,
            )
            return 1

    def _call_shell(
        self,
        args: [],
        manifest: "Manifest",
        cmd: str,
        step: str,
        tag: "Tag",
        shell: bool,
    ):
        env = dict(os.environ)
        env["HAMSTERCAGE_CMD"] = cmd
        env["HAMSTERCAGE_MANIFEST"] = manifest.manifest_file
        env["HAMSTERCAGE_HOOK"] = self.name
        env["HAMSTERCAGE_REPO"] = str(Path(manifest.manifest_file).parent)
        env["HAMSTERCAGE_STEP"] = step
        env["HAMSTERCAGE_TAG"] = tag.name
        r = subprocess.call(args, env=env, shell=shell)
        if r != 0:
            raise HamstercageException(
                f'Error executing hook "{self.name}" "{" ".join(args)}": command exited with {r}',
                r,
            )
        return 0

    def _get_path(self, manifest: "Manifest"):
        path = Path(self.command)
        if not path.is_absolute():
            path = Path(manifest.manifest_file).parent / path
        return path


class Tag:
    """
    Represents a tag definition in the manifest.
    """

    description: str
    name: str

    def __init__(self, name: str, description=None) -> None:
        self.name = name
        self.description = ""
        if description is not None:
            self.description = description
        self.entries = {}
        self.hooks = {}

    @staticmethod
    def from_dict(name: str, d: dict) -> "Tag":
        tag = Tag(name)
        if "description" in d:
            tag.description = d["description"]
        if "entries" in d:
            for p, e in d["entries"].items():
                entry = Entry.entry(p, e)
                tag.entries[p] = entry
        if "hooks" in d:
            for p, e in d["hooks"].items():
                entry = Hook.from_dict(p, e)
                tag.hooks[p] = entry
        return tag

    def to_dict(self) -> dict:
        d = {}
        if len(self.description) > 0:
            d["description"] = self.description
        if len(self.entries) > 0:
            d["entries"] = {}
            for path, entry in self.entries.items():
                d["entries"][path] = entry.to_dict()
        if len(self.hooks) > 0:
            d["hooks"] = {}
            for path, entry in self.hooks.items():
                d["hooks"][path] = entry.to_dict()
        return d

    def find_hook(self, command: str, step: str) -> Hook:
        """
        Find the best match hook for this command and step. If no such hook is defined, return None.
        :param command:
        :param step:
        :return:
        """
        hook = None
        for n in (f"{step}-{command}", f"*-{command}", f"{step}-*", "*"):
            hook = self.hooks.get(n)
            if hook:
                break
        return hook

    def __str__(self):
        return f"<Tag name={self.name} entries={len(self.entries)}>"


class Manifest:
    dir_mode: int
    file_mode: int
    group: int
    hosts: dict
    manifest_file: str
    owner: int
    tags: dict

    def __init__(self, file: str) -> None:
        self.manifest_file = str(file)
        self.hosts = {}
        self.tags = {}

    def load(self) -> None:
        """
        Load the manifest from the repository.
        :return:
        """
        with open(self.manifest_file, "r") as stream:
            manifest = yaml.safe_load(stream)
        mp = Path(self.manifest_file)
        mst = mp.stat()
        self.owner = mst.st_uid
        self.group = mst.st_gid
        self.file_mode = mst.st_mode
        self.dir_mode = (
            self.file_mode | (self.file_mode & 0o0444) >> 2
        )  # copy r bit to x bit
        self.hosts = {}
        self.tags = {}
        for name, tag in manifest["tags"].items():
            self.tags[name] = Tag.from_dict(name, tag)
        for n, h in manifest["hosts"].items():
            host = Host.from_dict(n, h)
            self.hosts[n] = host

    def dump(self) -> None:
        """
        Save the manifest to the repository.
        :return:
        """
        manifest = {
            "hosts": {},
            "tags": {},
        }
        for name, entry in self.hosts.items():
            manifest["hosts"][name] = entry.to_dict()
        for name, entry in self.tags.items():
            manifest["tags"][name] = entry.to_dict()
        with open(self.manifest_file, "w") as stream:
            yaml.dump(manifest, stream)

    @staticmethod
    def normalize_path(path):
        """
        Returns the normal form of a path
        :return:
        """
        path = str(path)
        if path.endswith("/"):
            path = path[0:-1]
        return path



class HamstercageException(Exception):
    """
    An application runtime error that leads to exiting the application
    """

    def __init__(self, msg, exit_code=1):
        super().__init__(msg)
        self.exit_code = exit_code
        self.msg = msg



import io
import os
import unittest
from datetime import datetime
from pathlib import Path
from unittest import TestCase

import grp
import pytest
import time
from importlib_resources import files
from pwd import getpwuid

from hamstercage.__main__ import Hamstercage
from hamstercage.hamstercage_exception import HamstercageException
from hamstercage.manifest import Hook
from hamstercage.utils import chmod

RUNNING_ON_GITHUB = int(os.environ.get("RUNNING_ON_GITHUB", 0))


class TestHamstercage(TestCase):
    @pytest.fixture(autouse=True)
    def initdir(self, tmpdir):
        self.tmpdir = tmpdir
        self.user = getpwuid(os.stat(self.tmpdir).st_uid).pw_name
        self.group = grp.getgrgid(os.stat(self.tmpdir).st_gid).gr_name
        self.now = time.time()

    def prepare_hamstercage(self, create=True) -> Hamstercage:
        dut = Hamstercage()
        dut.manifest_file = self.tmpdir / "hamstercage.yaml"
        dut.hostname = "testing.example.com"
        dut.target = Path(self.tmpdir) / "target"
        dut.repo = Path(self.tmpdir) / "repo"
        if create:
            dut.init(None)
            chmod(dut.manifest_file, 0o664)
            dut.target.mkdir()
        return dut

    def _add_file(self, files: dict, name: str, target: Path):
        files[name] = target / name
        files[name].write_text(f"Test file {name}")

    def test_load(self):
        dut = Hamstercage()
        dut.manifest_file = files("hamstercage.tests") / "hamstercage.yaml"
        dut.hostname = "testing.example.com"
        dut._load_manifest()

        self.assertIn("testing.example.com", dut.manifest.hosts)
        host = dut.manifest.hosts["testing.example.com"]
        self.assertIn("all", host.tags)

        self.assertIn("all", dut.manifest.tags)
        tag = dut.manifest.tags["all"]
        self.assertEqual(tag.description, "files that apply to all hosts")
        self.assertIn("foo.txt", tag.entries)

    def test_add_dir(self):
        dut = self.prepare_hamstercage()

        dir_to_add = "a-dir"
        path_to_add = dut.target / dir_to_add
        path_to_add.mkdir()
        os.utime(path_to_add, (self.now, self.now))

        args = Args(files=[dir_to_add], tag="all")
        r = dut.add(args)
        self.assertEqual(0, r)

        m = dut.manifest_file.read_text("utf-8")
        self.assertEqual(
            (
                "hosts:\n"
                "  testing.example.com:\n"
                "    tags:\n"
                "    - all\n"
                "tags:\n"
                "  all:\n"
                "    description: files that apply to all hosts\n"
                "    entries:\n"
                "      /a-dir:\n"
                "        group: " + self.group + "\n"
                "        mode: 0o755\n"
                "        owner: " + self.user + "\n"
                "        type: dir\n"
            ),
            m,
        )

        self.assertFalse((dut.repo / "tags" / "all" / dir_to_add).exists())

    def test_add_file(self):
        dut = self.prepare_hamstercage()

        file_to_add = "foo.txt"
        path_to_add = dut.target / file_to_add
        path_to_add.write_text("Hello, world!", "utf-8")
        path_to_add.chmod(0o760)
        os.utime(path_to_add, (self.now, self.now))

        args = Args(files=[path_to_add], tag="all")
        r = dut.add(args)
        self.assertEqual(0, r)

        m = dut.manifest_file.read_text("utf-8")
        self.assertEqual(
            (
                "hosts:\n"
                "  testing.example.com:\n"
                "    tags:\n"
                "    - all\n"
                "tags:\n"
                "  all:\n"
                "    description: files that apply to all hosts\n"
                "    entries:\n"
                "      /foo.txt:\n"
                "        group: " + self.group + "\n"
                "        mode: 0o760\n"
                "        owner: " + self.user + "\n"
                "        type: file\n"
            ),
            m,
        )

        f = (dut.repo / "tags" / "all" / file_to_add).read_text("utf-8")
        self.assertEqual("Hello, world!", f)

    def test_add_link(self):
        dut = self.prepare_hamstercage()

        link_to_add = "a-link"
        path_to_add = dut.target / link_to_add
        path_to_add.symlink_to("/dev/null")
        os.utime(path_to_add, (self.now, self.now), follow_symlinks=False)

        args = Args(files=[link_to_add], tag="all")
        r = dut.add(args)
        self.assertEqual(0, r)

        m = dut.manifest_file.read_text("utf-8")
        self.assertEqual(
            (
                "hosts:\n"
                "  testing.example.com:\n"
                "    tags:\n"
                "    - all\n"
                "tags:\n"
                "  all:\n"
                "    description: files that apply to all hosts\n"
                "    entries:\n"
                "      /a-link:\n"
                "        target: /dev/null\n"
                "        type: link\n"
            ),
            m,
        )

        self.assertFalse((dut.repo / "tags" / "all" / link_to_add).exists())

    def test_add_many(self) -> Hamstercage:
        self.perform_add_many()

    def perform_add_many(self) -> Hamstercage:
        dut = self.prepare_hamstercage()

        self.dir_to_add = "a-dir"
        self.dir_path = dut.target / self.dir_to_add
        self.dir_path.mkdir()
        os.utime(self.dir_path, (self.now, self.now))

        self.file_to_add = "foo.txt"
        self.file_path = dut.target / self.file_to_add
        self.file_path.write_text("Hello, world!", "utf-8")
        self.file_path.chmod(0o644)
        os.utime(self.file_path, (self.now, self.now))

        self.link_to_add = "a-link"
        self.link_path = dut.target / self.link_to_add
        self.link_path.symlink_to("/dev/null")
        os.utime(self.link_path, (self.now, self.now), follow_symlinks=False)

        args = Args(
            files=[self.dir_to_add, self.file_to_add, self.link_to_add], tag="all"
        )
        r = dut.add(args)
        self.assertEqual(0, r)

        return dut

    def test_apply(self):
        self.perform_apply()

    def perform_apply(self):
        dut = self.perform_add_many()
        hook_status_file = self.tmpdir / "hook-ran"
        hook = Hook.from_dict(
            "post-apply",
            {"command": f"echo foo >{hook_status_file}", "type": "shell"},
        )
        dut.manifest.tags["all"].hooks[hook.name] = hook

        dut.target = Path(self.tmpdir) / "apply"
        args = Args(files=[])
        r = dut.apply(args)
        self.assertEqual(0, r)
        self.assert_path_equal(self.dir_path, dut.target / self.dir_to_add)
        self.assert_path_equal(self.file_path, dut.target / self.file_to_add)
        self.assert_path_equal(self.link_path, dut.target / self.link_to_add)
        assert hook_status_file.exists()

        # again to make sure it's idempotent
        r = dut.apply(args)
        self.assertEqual(0, r)
        self.assert_path_equal(self.dir_path, dut.target / self.dir_to_add)
        self.assert_path_equal(self.file_path, dut.target / self.file_to_add)
        self.assert_path_equal(self.link_path, dut.target / self.link_to_add)

        return dut

    @unittest.skipIf(
        RUNNING_ON_GITHUB, "Github runner does not support mode bits on /tmp"
    )
    def test_apply_mode_change(self):
        self.perform_apply_mode_change()

    def perform_apply_mode_change(self):
        dut = self.perform_add_many()
        hook_status_file = self.tmpdir / "hook-ran"
        hook = Hook.from_dict(
            "post-apply",
            {"command": f"echo foo >{hook_status_file}", "type": "shell"},
        )
        dut.manifest.tags["all"].hooks[hook.name] = hook

        dut.target = Path(self.tmpdir) / "apply"
        args = Args(files=[])
        r = dut.apply(args)
        self.assertEqual(0, r)
        self.assert_path_equal(self.dir_path, dut.target / self.dir_to_add)
        self.assert_path_equal(self.file_path, dut.target / self.file_to_add)
        self.assertEqual((dut.target / self.file_to_add).stat().st_mode & 0o7777, 0o644)
        self.assert_path_equal(self.link_path, dut.target / self.link_to_add)
        assert hook_status_file.exists()

        dut.manifest.tags["all"].entries["/foo.txt"].mode = 0o444

        r = dut.apply(args)
        self.assertEqual(0, r)
        self.assertEqual((dut.target / self.file_to_add).stat().st_mode & 0o7777, 0o444)

        return dut

    def test_apply_target_exists(self):
        dut = self.perform_apply()

        dir_to_add = "a-dir"
        dir_path = dut.target / dir_to_add
        file_to_add = "foo.txt"
        file_path = dut.target / file_to_add
        link_to_add = "a-link"
        link_path = dut.target / link_to_add

        dir_path.rename(dut.target / "temp")
        file_path.rename(dut.target / "a-dir")
        link_path.rename(dut.target / "foo.txt")
        (dut.target / "temp").rename(dut.target / "a-link")

        args = Args(files=[])
        with self.assertRaises(HamstercageException):
            r = dut.apply(args)

    def test_diff_unchanged(self):
        dut = self.perform_add_many()

        args = Args(files=[])
        r = dut.diff(args)
        self.assertEqual(0, r)

    def test_diff_changed(self):
        dut = self.perform_add_many()

        self.file_path.write_text("Goodbye, world!", "utf-8")

        args = Args(files=[])
        r = dut.diff(args)
        self.assertEqual(1, r)

    def test_diff_missing(self):
        dut = self.perform_add_many()
        self.file_path.unlink()

        args = Args(files=[])
        r = dut.diff(args)
        self.assertEqual(1, r)

    def test_entries(self):
        dut = self.perform_add_many()

        entries = {}
        for t, e in dut._entries():
            entries[e.path] = (t, e)
        assert len(entries) == 3
        assert "/foo.txt" in entries
        assert "/a-dir" in entries
        assert "/a-link" in entries

    def test_entries_duplicate(self):
        dut = self.prepare_hamstercage()
        dut.tags = ["other", "all"]

        file_to_add = "foo.txt"
        path_to_add = dut.target / file_to_add
        path_to_add.write_text("Hello, world!", "utf-8")
        path_to_add.chmod(0o760)
        os.utime(path_to_add, (self.now, self.now))

        args = Args(name="other")
        r = dut.tag_add(args)
        self.assertEqual(0, r)

        args = Args(files=[file_to_add], tag="all")
        r = dut.add(args)
        self.assertEqual(0, r)

        args = Args(files=[file_to_add], tag="other")
        r = dut.add(args)
        self.assertEqual(0, r)

        entries = {}
        for t, e in dut._entries():
            entries[e.path] = (t, e)
        assert len(entries) == 1
        assert "/foo.txt" in entries
        assert entries["/foo.txt"][0] == "other"

    def test_init(self):
        dut = self.prepare_hamstercage()

        m = dut.manifest_file.read_text("utf-8")
        self.assertEqual(
            (
                "hosts:\n"
                "  testing.example.com:\n"
                "    tags:\n"
                "    - all\n"
                "tags:\n"
                "  all:\n"
                "    description: files that apply to all hosts\n"
            ),
            m,
        )

    def test_list_short(self):
        dut = self.perform_add_many()
        out = io.StringIO()

        args = Args(files=[], long=0)
        r = dut.list(args, file=out)
        self.assertEqual(0, r)

        t = datetime.fromtimestamp(os.stat(self.file_path).st_mtime).strftime("%H:%M")
        self.assertEqual(
            [
                str(self.dir_path),
                str(self.link_path),
                str(self.file_path),
                "",
            ],
            out.getvalue().split("\n"),
        )

    def test_list_long(self):
        dut = self.perform_add_many()
        out = io.StringIO()

        args = Args(files=[], long=1)
        r = dut.list(args, file=out)
        self.assertEqual(0, r)

        ts_dir = datetime.fromtimestamp(os.stat(self.dir_path).st_mtime).strftime(
            "%H:%M"
        )
        ts_file = datetime.fromtimestamp(os.stat(self.file_path).st_mtime).strftime(
            "%H:%M"
        )
        ts_link = datetime.fromtimestamp(os.stat(self.link_path).st_mtime).strftime(
            "%H:%M"
        )
        self.assertEqual(
            [
                f" \tdrwxr-xr-x\t{self.user}\t{self.group}\t0\t{ts_dir}\tall\t{self.dir_path}/",
                f" \tlrw-r--r--\troot\troot\t0\t{ts_link}\tall\t{self.link_path} -> /dev/null",
                f" \t-rw-r--r--\t{self.user}\t{self.group}\t13\t{ts_file}\tall\t{self.file_path}",
                "",
            ],
            out.getvalue().split("\n"),
        )

    def test_list_long_one(self):
        dut = self.perform_add_many()
        out = io.StringIO()

        args = Args(files=[str(self.file_path)], long=1)
        r = dut.list(args, file=out)
        self.assertEqual(0, r)

        ts_file = datetime.fromtimestamp(os.stat(self.file_path).st_mtime).strftime(
            "%H:%M"
        )
        self.assertEqual(
            [
                f" \t-rw-r--r--\t{self.user}\t{self.group}\t13\t{ts_file}\tall\t{self.file_path}",
                "",
            ],
            out.getvalue().split("\n"),
        )

    def test_list_long_missing(self):
        dut = self.perform_add_many()
        out = io.StringIO()
        self.file_path.unlink()

        args = Args(files=[str(self.file_path)], long=1)
        r = dut.list(args, file=out)
        self.assertEqual(0, r)

    def test_main(self):
        dut = self.prepare_hamstercage()
        dut.main([])

    def test_normalize_target_path(self):
        dut = self.prepare_hamstercage()

        assert dut._normalize_target_path("/foo") == ("/foo", dut.target / "foo")
        assert dut._normalize_target_path("foo") == ("/foo", dut.target / "foo")
        assert dut._normalize_target_path(dut.target / "foo") == (
            "/foo",
            dut.target / "foo",
        )

    def test_remove_file(self):
        dut = self.prepare_hamstercage()

        file_to_add = "foo.txt"
        path_to_add = dut.target / file_to_add
        path_to_add.write_text("Hello, world!", "utf-8")
        path_to_add.chmod(0o760)

        args = Args(files=["/" + file_to_add], tag="all")
        r = dut.add(args)
        self.assertEqual(0, r)

        m = dut.manifest_file.read_text("utf-8")
        self.assertEqual(
            (
                "hosts:\n"
                "  testing.example.com:\n"
                "    tags:\n"
                "    - all\n"
                "tags:\n"
                "  all:\n"
                "    description: files that apply to all hosts\n"
                "    entries:\n"
                "      /foo.txt:\n"
                "        group: " + self.group + "\n"
                "        mode: 0o760\n"
                "        owner: " + self.user + "\n"
                "        type: file\n"
            ),
            m,
        )

        f = (dut.repo / "tags" / "all" / file_to_add).read_text("utf-8")
        self.assertEqual("Hello, world!", f)

        r = dut.remove(args)
        self.assertEqual(0, r)
        self.assertFalse((dut.repo / "tags" / "all" / file_to_add).exists())

    def test_save(self):
        dut = self.perform_add_many()
        dut.tags = ["other", "all"]
        dut.manifest.dump()

        dir_to_add = "a-dir"
        dir_path = dut.target / dir_to_add
        file_to_add = "foo.txt"
        file_path = dut.target / file_to_add
        link_to_add = "a-link"
        link_path = dut.target / link_to_add

        dir_path.chmod(0o700)
        file_path.write_text("Foo bar", "utf-8")
        link_path.unlink()
        link_path.symlink_to("/dev/zero")

        dut = self.prepare_hamstercage(create=False)
        r = dut.save(Args(force=0))
        self.assertEqual(0, r)

        entry = dut.manifest.tags["all"].entries["/" + dir_to_add]
        self.assertEqual(0o700, entry.mode)
        entry = dut.manifest.tags["all"].entries["/" + file_to_add]
        self.assertEqual("Foo bar", dut._path_repo_entry("all", entry).read_text())
        entry = dut.manifest.tags["all"].entries["/" + link_to_add]
        self.assertEqual("/dev/zero", entry.target)

    def test_save_duplicate(self):
        dut = self.prepare_hamstercage()

        args = Args(name="other")
        r = dut.tag_add(args)
        self.assertEqual(0, r)

        files = {}

        self._add_file(files, "only_in_all.txt", dut.target)
        self._add_file(files, "only_in_other.txt", dut.target)

        args = Args(files=[files["only_in_all.txt"]], tag="all")
        r = dut.add(args)
        self.assertEqual(0, r)

        args = Args(files=[files["only_in_other.txt"]], tag="other")
        r = dut.add(args)
        self.assertEqual(0, r)

        assert "/only_in_all.txt" in dut.manifest.tags["all"].entries
        assert (dut.repo / "tags" / "all" / "only_in_all.txt").exists()
        assert not (dut.repo / "tags" / "all" / "only_in_other.txt").exists()
        assert "/only_in_other.txt" in dut.manifest.tags["other"].entries
        assert (dut.repo / "tags" / "other" / "only_in_other.txt").exists()
        assert not (dut.repo / "tags" / "other" / "only_in_all.txt").exists()

        args = Args()
        r = dut.save(args)
        self.assertEqual(0, r)

        assert "/only_in_all.txt" in dut.manifest.tags["all"].entries
        assert (dut.repo / "tags" / "all" / "only_in_all.txt").exists()
        assert not (dut.repo / "tags" / "all" / "only_in_other.txt").exists()
        assert "/only_in_other.txt" in dut.manifest.tags["other"].entries
        assert (dut.repo / "tags" / "other" / "only_in_other.txt").exists()
        assert not (dut.repo / "tags" / "other" / "only_in_all.txt").exists()

    def test_tag_add(self):
        dut = self.prepare_hamstercage()
        dut._load_manifest()

        args = Args(name="other")
        r = dut.tag_add(args)
        self.assertEqual(0, r)

        assert "other" in dut.manifest.tags
        assert dut.manifest.tags["other"].description == ""

        args = Args(name="other")
        with self.assertRaises(HamstercageException):
            dut.tag_add(args)

        args = Args(name="foo", description="bar")
        r = dut.tag_add(args)
        self.assertEqual(0, r)

        assert "foo" in dut.manifest.tags
        assert dut.manifest.tags["foo"].description == "bar"

    def assert_path_equal(self, expected_path: Path, actual_path: Path):
        if not expected_path.exists():
            raise AssertionError(f"{expected_path} does not exist")
        if not actual_path.exists():
            raise AssertionError(f"{actual_path} does not exist")
        expected_stat = expected_path.stat()
        actual_stat = actual_path.stat()
        if expected_path.is_dir() and not actual_path.is_dir():
            raise AssertionError(f"{expected_path} should be a directory but isn't")
        if expected_path.is_file() and not actual_path.is_file():
            raise AssertionError(f"{expected_path} should be a file but isn't")
        if expected_path.is_symlink() and not actual_path.is_symlink():
            raise AssertionError(f"{expected_path} should be a symlink but isn't")
        if expected_stat.st_uid != actual_stat.st_uid:
            raise AssertionError(
                f"expected owner {expected_stat.st_uid} but found {actual_stat.st_uid}"
            )
        if expected_stat.st_gid != actual_stat.st_gid:
            raise AssertionError(
                f"expected group {expected_stat.st_gid} but found {actual_stat.st_gid}"
            )
        if expected_stat.st_mode & 0o7777 != actual_stat.st_mode & 0o7777:
            raise AssertionError(
                f"expected {expected_stat.st_mode:o}, but was {actual_stat.st_mode:o}"
            )


class Args:
    files = []
    force = 0

    def __init__(
        self, description=None, files=None, force=0, long=0, name=None, tag=None
    ):
        if files is None:
            files = []
        self.description = description
        self.files = files
        self.force = force
        self.long = long
        self.name = name
        self.tag = tag



import io
import unittest
from datetime import datetime, timedelta

from hamstercage.utils import mode_to_str, short_date, print_table


class TestHamstercage(unittest.TestCase):
    def test_mode_to_str_ug_rw(self):
        assert mode_to_str("-", 0o660) == "-rw-rw----"

    def test_mode_to_str_one_each(self):
        assert mode_to_str("-", 0o124) == "---x-w-r--"

    def test_mode_to_str_ug_rw_us(self):
        assert mode_to_str("-", 0o4660) == "-rwSrw----"

    def test_mode_to_str_ug_rwx_gs(self):
        assert mode_to_str("-", 0o2770) == "-rwxrws---"

    def test_mode_to_str_ug_rwx_os(self):
        assert mode_to_str("-", 0o1770) == "-rwxrwx--T"

    def test_print_table_with_tabs(self):
        file = io.StringIO()

        print_table([["1.1", "1.2"], ["2.1", "2.2"]], file=file)

        assert file.getvalue() == "1.1\t1.2\n2.1\t2.2\n"

    def test_print_table_without_tabs(self):
        file = io.StringIO()

        print_table(
            [["1.1 long", "1.2 long"], ["2.1", "2.2"]],
            align=["<", ">"],
            file=file,
            tabs=False,
        )

        assert file.getvalue() == "1.1 long 1.2 long\n2.1      2.2\n"

    def test_short_date_now(self):
        t = datetime.now()
        assert short_date(int(t.timestamp())) == t.strftime("%H:%M")

    def test_short_date_yesterday(self):
        t = datetime.now() + timedelta(days=-1, seconds=-2)
        assert short_date(int(t.timestamp())) == t.strftime("%d.%m.")

    def test_short_date_9_months_ago(self):
        t = datetime.now() + timedelta(days=-9 * 30)
        assert short_date(int(t.timestamp())) == t.strftime("%Y")



import argparse
import os
import shutil
import socket
from datetime import datetime
from difflib import unified_diff
from pathlib import Path
from typing import Iterator, Union

import sys
from yaml.scanner import ScannerError

from hamstercage import Manifest
from hamstercage.hamstercage_exception import HamstercageException
from hamstercage.manifest import Host, Tag, Entry, DirEntry, SymlinkEntry, FileEntry
from hamstercage.utils import (
    chmod,
    mode_to_str,
    short_date,
    print_table,
    ListEntry,
    path_as_child_of,
)


class Hamstercage:
    """
    The main progam. Parses command line arguments and invokes sub-commands.
    """

    target: Path
    manifest_file: Path
    files: list
    hostname: str
    repo: Path
    tags: list

    def __init__(self):
        self.files = []
        self.hostname = socket.gethostname()
        self.manifest = None
        self.manifest_file = Path("hamstercage.yaml")
        self.repo = Path(".")
        self.tags = []
        self.target = Path("/")

    def main(self, args=None):
        if args is None:
            args = sys.argv[1:]
        parser = argparse.ArgumentParser(
            prog="hamstercage", description="Manage the hamster cage."
        )
        parser.add_argument(
            "-d",
            "--directory",
            type=Path,
            default="/",
            help="base directory of target files",
        )
        parser.add_argument(
            "-f",
            "--file",
            type=Path,
            default="hamstercage.yaml",
            help="manifest file to use",
        )
        parser.add_argument(
            "-n", "--hostname", default=socket.gethostname(), help="name of this host"
        )
        parser.add_argument(
            "-r", "--repo", type=Path, default=".", help="directory of file repo"
        )
        parser.add_argument("-t", "--tag", type=str, help="tags to apply/save")
        parser.add_argument(
            "-v", "--verbose", action="count", default=0, help="verbose output"
        )
        parser.set_defaults(func=None)

        subparsers = parser.add_subparsers(help="sub-command help")

        subparser = subparsers.add_parser(
            "add", help="add one or more files to the manifest"
        )
        subparser.set_defaults(func=self.add)
        subparser.add_argument(
            "-f",
            "--force",
            action="count",
            default=0,
            help="overwrite existing entries",
        )
        subparser.add_argument("tag", help="tag to add files to")
        subparser.add_argument("files", nargs="+", help="files to add")

        subparser = subparsers.add_parser(
            "apply", help="apply files from repo to target"
        )
        subparser.set_defaults(func=self.apply)
        subparser.add_argument(
            "files", nargs="*", help="limit results to these file patterns"
        )

        subparser = subparsers.add_parser(
            "diff", help="print differences between target and repo"
        )
        subparser.set_defaults(func=self.diff)
        subparser.add_argument(
            "files", nargs="*", help="limit results to these file patterns"
        )

        subparser = subparsers.add_parser("init", help="create a new manifest")
        subparser.set_defaults(func=self.init)

        subparser = subparsers.add_parser(
            "list", aliases=["ls"], help="list manifest entries"
        )
        subparser.set_defaults(func=self.list)
        subparser.add_argument(
            "-l", "--long", action="count", default=0, help="list format long"
        )
        subparser.add_argument(
            "-t",
            "--tabs",
            action="count",
            default=0,
            help="separate columns with tabs instead of spaces",
        )
        subparser.add_argument(
            "files", nargs="*", help="limit results to these file patterns"
        )

        subparser = subparsers.add_parser(
            "remove",
            aliases=["del", "rm"],
            help="remove one or more files from the manifest",
        )
        subparser.set_defaults(func=self.remove)
        subparser.add_argument("tag", help="tag to add files to")
        subparser.add_argument("files", nargs="+", help="files to remove")

        subparser = subparsers.add_parser("save", help="save target files to repo")
        subparser.set_defaults(func=self.save)
        subparser.add_argument(
            "files", nargs="*", help="limit results to these file patterns"
        )

        subparser = subparsers.add_parser(
            "tag",
            help="manage tags in the manifest",
        )
        subparser.set_defaults(func=None)

        tagparsers = subparser.add_subparsers(help="tag command help")

        subparser = tagparsers.add_parser(
            "add",
            help="add a tag to the manifest",
        )
        subparser.set_defaults(func=self.tag_add)
        subparser.add_argument("name", help="tag name")
        subparser.add_argument(
            "-d",
            "--description",
            type=str,
            default="",
            help="description for this tag",
        )

        args = parser.parse_args(args)

        if args.func:
            self.target = args.directory
            self.manifest_file = args.file
            self.hostname = args.hostname
            self.repo = args.repo
            if args.tag:
                self.tags = [args.tag]

            try:
                return args.func(args)
            except HamstercageException as e:
                print(f"{e}", file=sys.stderr)
                return e.exit_code
        parser.print_help()
        return 64

    def add(self, args):
        """
        Add one or more entries to the manifest.
        :param args:
        :return:
        """
        self._load_manifest()
        if len(args.files) < 1:
            raise HamstercageException(f"Need at least one file to add", 64)

        self._run_hook(args.tag, "add", "pre")
        for file in args.files:
            entry = self._add_entry(args.tag, file, ignore_existing=args.force > 0)
            self._save_entry(args.tag, entry)
        self._run_hook(args.tag, "add", "post")
        self.manifest.dump()
        return 0

    def apply(self, args):
        """
        Copy (all/matching) manifest entries from the repo to the target.
        :return:
        """
        self._load_manifest()
        self._run_hooks("apply", "pre")
        for t, e in self._entries():
            e.apply(self._path_repo_entry(t, e), self._path_target(e))
        self._run_hooks("apply", "post")
        return 0

    def diff(self, args):
        self._load_manifest()
        has_diff = False
        self.files = args.files

        self._run_hooks("diff", "pre")
        for t, e in self._entries():
            repo = self._path_repo_entry(t, e)
            target = self._path_target(e)
            if not repo.is_file():
                continue  # non-files don't have a file under tags
            if target.exists() and repo.exists():
                diff = list(self._diff(target, repo))
                sys.stdout.writelines(diff)
                if diff:
                    has_diff = True
            else:
                if self._mtime_or_missing(repo, "---"):
                    has_diff = True
                if self._mtime_or_missing(target, "+++"):
                    has_diff = True
        self._run_hooks("diff", "post")
        return 1 if has_diff else 0

    def init(self, args) -> int:
        """
        Create a manifest file. Throw an error if the file already exists.
        :return:
        """
        if self.manifest_file.exists():
            raise HamstercageException(
                f'manifest file "{self.manifest_file}" already exists', 1
            )
        manifest = Manifest(str(self.manifest_file))
        manifest.hosts[self.hostname] = Host(self.hostname)
        manifest.hosts[self.hostname].tags = ["all"]
        manifest.tags = {
            "all": Tag.from_dict(
                "all", {"description": "files that apply to all hosts", "entries": {}}
            )
        }
        manifest.dump()
        return 0

    def list(self, args, file=None):
        """
        Print a list of all manifest entries
        :return:
        """
        if file is None:
            out = sys.stdout
        self._load_manifest()
        self.files = args.files
        items = {}
        for t, e in self._entries():
            repo = self._path_repo_entry(t, e)
            target = self._path_target(e)
            if target not in items:
                items[target] = ListEntry(e, repo, t)
        if args.long == 0:
            for path in sorted(items):
                print(path, file=file)
        else:
            lines = []
            for path in sorted(items):
                item = items[path]
                entry = item.entry
                mtime = "?"
                name = str(path)
                size = "0"
                status = " "
                type = "-"
                if path.exists():
                    stat = path.stat()
                    mtime = short_date(int(stat.st_mtime))
                    if path.is_file():
                        size = str(stat.st_size)
                else:
                    status = "!"
                if isinstance(entry, FileEntry):
                    if path.exists() and list(self._diff(path, item.repo)):
                        status = "*"
                elif isinstance(entry, DirEntry):
                    name = name + "/"
                    type = "d"
                elif isinstance(entry, SymlinkEntry):
                    name = name + " -> " + entry.target
                    type = "l"
                    if path.exists() and os.readlink(path) != entry.target:
                        status = "*"
                lines.append(
                    [
                        status,
                        mode_to_str(type, entry.mode),
                        entry.owner,
                        entry.group,
                        size,
                        mtime,
                        item.tag,
                        name,
                    ]
                )
            print_table(lines, align=["<", "<", "<", "<", ">"], file=file)
        # widths = [0] * 8
        # align = ["<", "<", "<", "<", ">", "<", "<", "<"]
        # for line in lines:
        #     for i in range(0, len(line)):
        #         widths[i] = max(widths[i], len(line[i]))
        # fs = []
        # for i in range(0, len(widths) - 1):
        #     fs.append(f"{{{i}:{align[i]}{widths[i]}}}")
        #     # f = f + f"{{0:{widths[i]}.{widths[i]}s}}"
        # fs.append(f"{{{len(widths)-1}}}")
        # for line in lines:
        #     # print("{0:} {1:}".format(*line), file=out)
        #     print(" ".join(fs).format(*line), file=out)
        return 0

    def remove(self, args):
        """
        Remove one or more entries from the manifest.
        :param args:
        :return:
        """
        self._load_manifest()
        if len(args.files) < 1:
            raise HamstercageException(f"Need at least one file to remove", 64)

        for file in args.files:
            if file in self.manifest.tags[args.tag].entries:
                repo = self._path_repo_entry(
                    args.tag, self.manifest.tags[args.tag].entries[file]
                )
                del self.manifest.tags[args.tag].entries[file]
                repo.unlink(missing_ok=True)
            else:
                print(
                    f"Unable to remove {file}: no such entry in tag {args.tag}",
                    file=sys.stderr,
                )
                return 71
        self.manifest.dump()
        return 0

    def save(self, args):
        """
        Copy (all/matching) manifest entries from the target to the repo.
        :return:
        """
        self._load_manifest()
        self._run_hooks("save", "pre")
        for tag, entry in self._entries():
            self._save_entry(tag, entry)
        self._run_hooks("save", "post")
        return 0

    def tag_add(self, args):
        self._load_manifest()
        if args.name in self.manifest.tags:
            raise HamstercageException(f"A tag named f{args.name} already exists")
        self.manifest.tags[args.name] = Tag(args.name, args.description)
        self.manifest.dump()
        return 0

    def _add_entry(
        self,
        tag: str,
        file,
        ignore_existing=False,
    ) -> Entry:
        """
        Add a new file to, or update its contents and attributes in the repo.

        By default, it is an error if the repo file exists, and no error if the repo file doesn't exist.

        :param tag: the tag to update
        :param ignore_existing: it is not an error if the file already exists in the repo (default False)
        :return: the updated entry
        """
        if tag not in self.manifest.tags:
            raise HamstercageException(f"no tag {tag} in manifest")
        entries = self.manifest.tags[tag].entries
        (repo_path, target_path) = self._normalize_target_path(file)
        entry = Entry.entry(repo_path, target_path)
        if not ignore_existing:
            if entry.path in entries:
                raise HamstercageException(
                    f"Unable to add {target_path}: already added to tag {tag}"
                )
        entries[repo_path] = entry
        return entry

    def _save_entry(self, tag: str, entry: Entry):
        (repo_path, target_path) = self._normalize_target_path(entry.path)
        entry.save(self._path_repo_path_tag(repo_path, tag), target_path, self.manifest)
        return entry

    @staticmethod
    def _diff(target, repo) -> Iterator[str]:
        """
        Generate a unified diff between the target and the repo file.

        :param target: the target file path
        :param repo: the repo file path
        :return: An iterator of diff lines
        """
        if not target.exists() or not repo.exists():
            r = []
            if not target.exists():
                r.append(f"! {target} not found")
            if not repo.exists():
                r.append(f"! {repo} not found")
            return r
        with open(target) as f:
            t = f.readlines()
        with open(repo) as f:
            r = f.readlines()
        return unified_diff(
            r,
            t,
            fromfile=str(repo),
            fromfiledate=Hamstercage._mtime(repo),
            tofile=str(target),
            tofiledate=Hamstercage._mtime(target),
        )

    def _entries(self):
        """
        Generator that produces all entries, potentially filtered by the list of files given on the command line. Only
        returns entries for the first match for a path.
        :return: (tag, entry)
        """
        paths = {}
        for t in self.tags:
            for p, e in self.manifest.tags[t].entries.items():
                if not self._files_match(e):
                    continue
                if p in paths:
                    continue
                paths[p] = True
                yield t, e

    @staticmethod
    def _exists_status(path) -> str:
        """
        Return a string representing whether the path exists or not.
        :param path: to check
        :return: a space if the file exists, an exclamation sign otherwise
        """
        return " " if path else "!"

    def _files_match(self, entry: Entry) -> bool:
        """
        Return if the entry matches the files given on the command line. If the list of files is empty, any entry will
        match.

        :param entry: to check
        :return: True if the list of files matches this entry
        """
        if len(self.files) == 0:
            return True
        for file in self.files:
            (repo_path, target_path) = self._normalize_target_path(file)
            if entry.path == repo_path:
                return True
        return False

    def _load_manifest(self) -> None:
        """
        Load the manifest from the configured path. If the manifest had been loaded previously, do nothing.
        :return: None
        """
        if self.manifest:
            return
        try:
            self.manifest = Manifest(str(self.manifest_file))
            self.manifest.load()
            if len(self.tags) == 0:
                if self.hostname in self.manifest.hosts:
                    self.tags = self.manifest.hosts[self.hostname].tags
                else:
                    print(
                        f"Warning: No hostname entry for {self.hostname}",
                        file=sys.stderr,
                    )
        except FileNotFoundError as e:
            raise HamstercageException(
                f'Unable to load manifest from "{self.manifest_file}": {e}', 71
            )
        except ScannerError as e:
            raise HamstercageException(
                f'Unable to load manifest from "{self.manifest_file}": {e}', 71
            )

    def _mkdir_repo(self, path: Path) -> None:
        """
        Create a directory in the repo, copying owner/group/mode from the manifest file
        :param path: of the new directory
        :return: None
        """
        if str(path) != ".":
            self._mkdir_repo(path.parent)
        p = self.repo.joinpath(path)
        if not p.exists():
            p.mkdir(mode=self.manifest.dir_mode)
            chmod(str(p), self.manifest.dir_mode)
            shutil.chown(str(p), self.manifest.owner, self.manifest.group)

    @staticmethod
    def _mtime(path: Path) -> str:
        """
        Return the modification time of path as a string.
        :param path: of file
        :return: time and date in ISO8601 format
        """
        return datetime.fromtimestamp(path.stat().st_mtime).isoformat()

    def _mtime_or_missing(self, path: Path, prefix: str):
        """
        Print the modification date of path, or "missing" if the file doesn't exist.
        :param path: of the file
        :param prefix: string prefix to print
        :return:
        """
        if path.exists():
            print(f"{prefix} {str(path)}\t{self._mtime(path)}")
            return False
        print(f"{prefix} {str(path)}\tmissing")
        return True

    def _normalize_target_path(self, path: Union[str, Path]) -> (str, Path):
        """
        Returns two paths: the path as it should be entered into the manifest, and the filesystem path for the target
        file.
        :param path:
        :return: list of manifest path (str), and filesystem path (Path)
        """
        ppath = Path(path)
        path = str(path)
        starget = str(self.target) + "/"
        if path.startswith(starget):
            # absolute path including target, strip target
            return path[len(starget) - 1 :], ppath.absolute()
        else:
            if ppath.is_absolute():
                # prepend target to path
                return str(path), Path(str(self.target) + str(path))
            else:
                #
                return "/" + str(path), Path(str(self.target) + "/" + str(path))

    def _path_repo_entry(self, tag: str, entry: Entry) -> Path:
        """
        Return the absolute path for the file of the entry.
        :param tag: the tag for this entry
        :param entry: of the file
        :return: path of file
        """
        return entry.path_as_child_of(self.repo / "tags" / tag)

    def _path_repo_path_tag(self, repo_path, tag: str):
        return path_as_child_of(repo_path, self.repo / "tags" / tag)

    def _path_target(self, entry: Entry) -> Path:
        """
        Return the path to the target file for the entry.
        :param entry: of the file
        :return: path of file
        """
        return entry.path_as_child_of(self.target)

    def _run_hook(self, tagname: str, cmd: str, step: str):
        """
        Execute the defined hook (if any) for the given command and step. If the hook is defined, but cannot be
        executed successfully, a HamstercageException is thrown.
        :param tagname: where the hook is defined
        :param cmd: command that is being executed
        :param step: pre or post
        :return: 0 if successful, any other exit code on failure.
        """
        if not tagname in self.manifest.tags:
            raise HamstercageException(f"Unknown tag {tagname}")
        tag = self.manifest.tags[tagname]
        hook = tag.find_hook(cmd, step)
        if hook:
            return hook.call(self.manifest, cmd, step, tag)
        return 0

    def _run_hooks(self, cmd: str, step: str):
        """
        Execute the defined hook (if any) for the given command and step for all tags. If the hook is defined, but
        cannot be executed successfully, a HamstercageException is thrown.
        :param cmd: command that is being executed
        :param step: pre or post
        :return: 0 if successful, any other exit code on failure.
        """
        for t in self.tags:
            r = self._run_hook(t, cmd, step)
            if r != 0:
                return r
        return 0


def main():
    h = Hamstercage()
    sys.exit(h.main())


if __name__ == "__main__":
    main()



import yaml

from hamstercage.manifest import Manifest, FileMode

__all__ = ["Manifest"]


def representer(dumper, data):
    return yaml.ScalarNode("tag:yaml.org,2002:int", oct(data))


yaml.add_representer(FileMode, representer)



import os
import shutil
import stat
from typing import Union

import sys

from datetime import datetime, timedelta
from pathlib import Path


def chmod(path, mode):
    """
    Change access mode on a target directory entry. Do not follow symlinks if the platform supports it.

    :param path: path to the file
    :param mode: mode to be set
    :return:
    """
    if os.chmod in os.supports_follow_symlinks:
        os.chmod(path, mode, follow_symlinks=False)
    else:
        pass


def mkdir_with_owner_group_mode(path: Path, owner: str, group: str, mode: int):
    """
    Create the directory path, including all parents. The new directories will be owned and have the permissions
    as specified.
    :param path: path of directory to create
    :param owner: owner of the new directory
    :param group: group of the new directory
    :param mode: permission bits of the new directory
    :return: none
    """
    if not path.exists():
        if path.parent is not None:
            mkdir_with_owner_group_mode(path.parent, owner, group, mode)
        path.mkdir(mode=mode)
        chmod(str(path), mode)
        shutil.chown(str(path), owner, group)


def mode_to_str(type: str, mode: int) -> str:
    """
    Converts the mode bits (from for example os.stat st_mode) into a string like ls.
    :param mode:
    :return:
    """
    return type + stat.filemode(mode)[1:]


def short_date(ts: int) -> str:
    """
    Returns a short date/time string based on the epoch timestamp
    :param ts:
    :return:
    """
    dt = datetime.fromtimestamp(ts)
    delta = abs(dt - datetime.now())
    f = "%Y"
    if delta < timedelta(days=1):
        f = "%H:%M"
    elif delta < timedelta(days=365 / 2):
        f = "%d.%m."
    return dt.strftime(f)


def path_as_child_of(path: Union[Path, str], target_path: Path) -> Path:
    """
    Returns the entry path as a child of target_path.
    :param path: the child path
    :param target_path: the base path
    :return: path
    """
    path = str(path)
    if path.startswith("/"):
        path = path[1:]
    return target_path / path


def print_table(table: list, align=None, file=None, tabs=None) -> None:
    """
    Prints the contents of the two-dimensional list.
    :param table:
    :param file:
    :return:
    """
    if align is None:
        align = []
    if file is None:
        file = sys.stdout
    if tabs is None:
        tabs = not file.isatty()
    cols = 0
    for line in table:
        cols = max(cols, len(line))
    if tabs:
        f = "\t".join(["{}"] * cols)
    else:
        widths = [0] * cols
        if len(align) < cols:
            align.extend(["<"] * (cols - len(align)))
        for line in table:
            for i in range(0, len(line)):
                widths[i] = max(widths[i], len(line[i]))
        fs = []
        for i in range(0, len(widths) - 1):
            fs.append(f"{{{i}:{align[i]}{widths[i]}}}")
            # f = f + f"{{0:{widths[i]}.{widths[i]}s}}"
        fs.append(f"{{{len(widths) - 1}}}")
        f = " ".join(fs)
    for line in table:
        # print("{0:} {1:}".format(*line), file=out)
        print(f.format(*line), file=file)


class ListEntry:
    # noinspection PyUnresolvedReferences
    def __init__(self, entry: "Entry", repo: Path, tag: str):
        self.entry = entry
        self.repo = repo
        self.tag = tag



import grp
import os
import pwd
import textwrap
import unittest
from pathlib import Path

import pytest
import yaml

from ..__main__ import Hamstercage
from ..hamstercage_exception import HamstercageException
from ..manifest import FileEntry, Manifest, Tag


class TestHamstercage(unittest.TestCase):
    @pytest.fixture(autouse=True)
    def initdir(self, tmpdir):
        self.tmpdir = tmpdir
        self.user = pwd.getpwuid(os.stat(self.tmpdir).st_uid).pw_name
        self.group = grp.getgrgid(os.stat(self.tmpdir).st_gid).gr_name

    def prepare_hamstercage(self) -> Hamstercage:
        dut = Hamstercage()
        dut.manifest_file = self.tmpdir / "hamstercage.yaml"
        dut.hostname = "testing.example.com"
        dut.target = Path(self.tmpdir) / "target"
        dut.repo = Path(self.tmpdir) / "repo"
        dut.init(None)
        os.chmod(dut.manifest_file, 0o664)
        dut.target.mkdir()
        return dut

    def manifest_with_hooks(self, hooks: dict) -> Manifest:
        manifest = {
            "hosts": {
                "testing.example.com": {"description": "testing hooks", "tags": ["all"]}
            },
            "tags": {
                "all": {
                    "description": "files that apply to all hosts",
                    "entries": {
                        "foo.txt": {
                            "group": self.group,
                            "mode": 0o644,
                            "owner": self.user,
                            "type": "file",
                        }
                    },
                    "hooks": hooks,
                }
            },
        }
        p = self.tmpdir / "with_hook.yaml"
        with open(p, "w") as stream:
            yaml.dump(manifest, stream)
        dut = Manifest(p)
        dut.load()
        return dut

    def test_FileEntry_from_dict(self):
        dut = FileEntry.from_dict(
            "foo",
            {
                "group": "wheel",
                "mode": 0o755,
                "owner": "root",
                "type": "file",
            },
        )
        assert dut.path == "foo"
        assert dut.group == "wheel"
        assert dut.mode == 0o755
        assert dut.owner == "root"

    def test_load_hook(self):
        dut = self.manifest_with_hooks(
            {
                "*": {
                    "command": "true",
                    "description": "a simple shell command",
                    "type": "shell",
                }
            }
        )

        assert "all" in dut.tags
        assert "*" in dut.tags["all"].hooks

        h = dut.tags["all"].hooks["*"]
        assert h.command == "true"
        assert h.description == "a simple shell command"
        assert h.type == "shell"

    def test_load_hook_invalid_type(self):
        with self.assertRaises(HamstercageException) as e:
            dut = self.manifest_with_hooks(
                {
                    "*": {
                        "command": "true",
                        "description": "a simple shell command",
                        "type": "foo",
                    }
                }
            )
        assert (
            e.exception.msg
            == 'In definition of hook "*": Invalid hook type "foo", must be one of exec, python, shell'
        )

    def hook_exec(self, command: str, script: Path):
        dut = self.manifest_with_hooks(
            {
                "*": {
                    "command": command,
                    "description": "a simple shell command",
                    "type": "exec",
                }
            }
        )
        assert "all" in dut.tags
        assert "*" in dut.tags["all"].hooks

        script.write_text(
            textwrap.dedent(
                f"""\
                #!/bin/sh
                set -e
                test "$1" = "apply"
                test "$2" = "post"
                test "$3" = "all"
                test "$HAMSTERCAGE_CMD" = "apply"
                test "$HAMSTERCAGE_HOOK" = "*"
                test "$HAMSTERCAGE_MANIFEST" = "{dut.manifest_file}"
                test "$HAMSTERCAGE_REPO" = "{str(Path(dut.manifest_file).parent)}"
                test "$HAMSTERCAGE_STEP" = "post"
                test "$HAMSTERCAGE_TAG" = "all"
                """
            ),
            "utf-8",
        )
        os.chmod(script, 0o755)

        h = dut.tags["all"].hooks["*"]
        r = h.call(dut, "apply", "post", dut.tags["all"])
        assert r == 0

    def test_run_hook_exec_absolute(self):
        self.hook_exec(str(self.tmpdir / "exec_test.sh"), self.tmpdir / "exec_test.sh")

    def test_run_hook_exec_relative(self):
        self.hook_exec("exec_test.sh", self.tmpdir / "exec_test.sh")

    def run_hook_python(self, python: str):
        script = self.tmpdir / "python_hook"
        dut = self.manifest_with_hooks(
            {
                "*": {
                    "command": "python_hook",
                    "description": "a simple shell command",
                    "type": "python",
                }
            }
        )
        assert "all" in dut.tags
        assert "*" in dut.tags["all"].hooks

        script.write_text(
            python,
            "utf-8",
        )
        return dut

    def test_run_hook_python_success(self):
        dut = self.run_hook_python(
            textwrap.dedent(
                f"""
                assert cmd == "apply"
                assert manifest is not None
                assert hook == "*"
                assert repo == "{self.tmpdir}"
                assert step == "post"
                assert tag.name == "all"
                print(__file__)
                assert __name__ == "__hamstercage__"
                """
            )
        )

        h = dut.tags["all"].hooks["*"]
        r = h.call(dut, "apply", "post", dut.tags["all"])
        assert r == 0

    def test_run_hook_python_syntax_error(self):
        dut = self.run_hook_python(
            textwrap.dedent(
                f"""
                in == valid
                """
            )
        )

        h = dut.tags["all"].hooks["*"]
        with self.assertRaises(HamstercageException) as e:
            r = h.call(dut, "apply", "post", dut.tags["all"])
        assert (
            e.exception.msg
            == 'Error executing hook "*" Python command "python_hook" line 2::\n\tin == valid'
        )

    def test_run_hook_shell_false(self):
        dut = self.manifest_with_hooks(
            {
                "*": {
                    "command": "true && false",
                    "description": "a simple shell command",
                    "type": "shell",
                }
            }
        )
        assert "all" in dut.tags
        assert "*" in dut.tags["all"].hooks

        h = dut.tags["all"].hooks["*"]
        with self.assertRaises(HamstercageException) as e:
            r = h.call(dut, "apply", "post", dut.tags["all"])
        assert (
            e.exception.msg
            == 'Error executing hook "*" "true && false": command exited with 1'
        )

    def test_run_hook_shell_true(self):
        dut = self.manifest_with_hooks(
            {
                "*": {
                    "command": "true && true",
                    "description": "a simple shell command",
                    "type": "shell",
                }
            }
        )
        assert "all" in dut.tags
        assert "*" in dut.tags["all"].hooks

        h = dut.tags["all"].hooks["*"]
        r = h.call(dut, "apply", "post", dut.tags["all"])
        assert r == 0

    def test_tag_find_hook_with_star(self):
        dut = Tag.from_dict(
            "all",
            {
                "hooks": {
                    "pre-add": {
                        "command": "pre-add",
                        "type": "shell",
                    },
                    "*-apply": {
                        "command": "star-apply",
                        "type": "shell",
                    },
                    "post-*": {
                        "command": "post-star",
                        "type": "shell",
                    },
                    "*": {
                        "command": "star",
                        "type": "shell",
                    },
                }
            },
        )

        h = dut.find_hook("add", "pre")
        assert h is not None and h.command == "pre-add"
        h = dut.find_hook("apply", "pre")
        assert h is not None and h.command == "star-apply"
        h = dut.find_hook("add", "post")
        assert h is not None and h.command == "post-star"
        h = dut.find_hook("list", "pre")
        assert h is not None and h.command == "star"

    def test_tag_find_hook_no_star(self):
        dut = Tag.from_dict(
            "all",
            {
                "hooks": {
                    "pre-add": {
                        "command": "pre-add",
                        "type": "shell",
                    },
                    "post-apply": {
                        "command": "post-apply",
                        "type": "shell",
                    },
                }
            },
        )

        h = dut.find_hook("add", "pre")
        assert h is not None and h.command == "pre-add"
        h = dut.find_hook("apply", "post")
        assert h is not None and h.command == "post-apply"
        h = dut.find_hook("add", "post")
        assert h is None
        h = dut.find_hook("list", "pre")
        assert h is None
